# -*- coding: utf-8 -*-

IMPLTYPE_POSTGRESQL = 'postgres'
IMPLTYPE_POSTGRESQL2 = 'postgres2'
IMPLTYPE_FIREBIRD = 'firebird'

CONF_HOST = 'host'
CONF_DB = 'db'
CONF_USER = 'user'
CONF_PASSW = 'password'

# sexos
SEXO_HOMBRE = 1
SEXO_MUJER = 2

# dias
LUNES = 1
MARTES = 2
MIERCOLES = 3
JUEVES = 4
VIERNES = 5
SABADO = 6
DOMINGO = 7

# privilegios
PRIVL_NINGUNO = 'NINGUNO'
PRIVL_LECTURA = 'READ-ONLY'
PRIVL_LECTURA_ESCRITURA = 'READ-WRITE'

# claves definidas para la definición de privilegios
CAMPOS = 'campos'
ETIQUETA = 'etiqueta'
ORDEN = 'orden'
UTILIZAR_SELECTOR = 'utilizar_selector'
BUSQUEDA = 'busqueda'
COD_CLASE = 'cod_clase'
ORDENACION = 'ordenacion'
CLASE = 'cls'

NOMBRE = 'nombre'
TABLA_RELACIONADA = 'tabla_relacionada'
TIPO = 'tipo'
VALOR = 'valor'
VALORES = 'valores'
SOLO_LECTURA = 'solo_lectura'
REQUERIDO = 'requerido'
TITULO = 'titulo'
CAMPOS_REFERENCIA = 'campos_referencia'

TYPE_DATE = 'date'
TYPE_TIME = 'time'
TYPE_BOOLEAN = 'boolean'
TYPE_STRING = 'string'
TYPE_CHAR = 'char'
TYPE_TEXT = 'text'
TYPE_INTEGER = 'integer'
TYPE_REAL = 'real'

# constantes para Firebird
VAR_LANZADOR = 'lanzador'
VAR_TIMEOUT_CONSULTAS = 'time_out_consultas'

#
STATUS = 'status'
MESSAGE = 'message'
EXPIRATION = 'expiration'

# idiomas
LANG_ES_ES = 0
LANG_EN_UK = 1
